package fr.wikimedia.searchEngine;

import java.io.*;
import java.util.regex.Pattern;
import java.util.*;
import java.util.zip.*;

class Indexer {
		ArrayList<Map<String, Set<Integer>>> mapping;
		ArrayList<String> filenames;
		Map<String, Integer> counts;
		Set<String> stopWords;

		int modulo;

		Indexer() {
				modulo = 1024;
				mapping = new ArrayList<Map<String, Set<Integer>>>(modulo);
				stopWords = new HashSet<String>();
				counts = new HashMap<String,Integer> ();
				for(int i=0; i<modulo; i++) mapping.add(new HashMap<String, Set<Integer>> ());
				filenames = new ArrayList<String>();
		}

		void add(String word, int index) {
				Integer countN = counts.get(word);
				int count;
				if (countN == null) {
						count = 1;
				} else {
						count = countN+1;
				}
				counts.put(word, count);

				if (stopWords.contains(word)) return;

				int hash = (word.hashCode() & 0x7FFFFFFF) % modulo;
				Map<String, Set<Integer>> submapping = mapping.get(hash);

				if (count > 10000) { // stop word
						System.out.println("stop word: " + word);
						stopWords.add(word);
						submapping.remove(word);
				} else {
						Set<Integer> set = submapping.get(word);
						if (set == null) {
								set = new TreeSet<Integer> ();
								submapping.put(word, set);
						}
						set.add(index);
				}
		}

		static void writeIndex(Object index, String filename) throws IOException {
				ObjectOutputStream oStream = new ObjectOutputStream(new GZIPOutputStream(new FileOutputStream(filename)));
				oStream.writeObject(index);
				oStream.close();
		}

		void writeTo(String prefix) throws IOException {
				writeIndex(filenames, prefix + "/article.idx");
				for(int i=0; i<modulo; i++) {
						writeIndex(mapping.get(i), prefix + "/word" + i + ".idx");
				}
		}

		public static void main(String[] args) throws IOException, org.xml.sax.SAXException {
				BufferedReader filenames = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
				Indexer indexer = new Indexer();
				HTMLToRawText converter = new HTMLToRawText();

				for(int index=0; ; index++) {
						String filename = filenames.readLine();
						indexer.filenames.add(index, filename);
						if (filename == null) break;
						System.out.println(index + " : " + filename);
						try {
								String text= converter.toRawText(new BufferedInputStream(new FileInputStream(filename)));
								Pattern pattern = Pattern.compile("[\\P{L}&&\\P{Nd}]+");
								for(String s : pattern.split(text)) {
										indexer.add(s.toLowerCase(), index);
								}
						}
						catch (IOException exn) {
								System.err.println(filename + " : " + exn);
						}
						catch (org.xml.sax.SAXException exn) {
								System.err.println(filename + " : " + exn);
						}
				}

				indexer.writeTo("da_index");
		}
}
